

/*****************************************************************/
/* bliksem.c is the main file of the bliksem theorem prover.     */
/* Bliksem 1.12 A                                                */
/*****************************************************************/


#define CLASS 
#include "basics.h"
#include "extfunc.h" 
#include "resolution.h"


#define SHOWFILTERS    0

/****************************************************************/
/*               Error functions:                               */
/****************************************************************/


/* Since the program is non-interactive, there is no recovery after */
/* an error.                                                        */

static void closefiles ( )
{
   int i;

   for( i = freeincludefiles - 1; i >= 0; i -- ) 
   {
      if( includefiles [i]. rw == FILE_READ )
      {
         fclose ( includefiles [i]. handle ); 
         includefiles [i]. rw = FILE_NOTINUSE;
      }   
   }
   if( inputfile. rw == FILE_READ )
   {
      fclose( inputfile. handle );
      inputfile. rw = FILE_NOTINUSE; 
   }
}

/* print s, followed by 'in line %d of file %f', and exit.   */

void errorwithline ( s )
char *s;
{
   int i;

   printf( "ERROR: %s ", s );
  
   for( i = freeincludefiles - 1; i >= 0; i -- )
   {
      if( includefiles [i]. rw == FILE_READ ) 
      {
         printf( "in line %d of file %s\n", 
                 includefiles [i]. linenum,
                 includefiles [i]. name );
      }
   }

   if( inputfile. rw == FILE_READ ) 
   {
      printf("in line %d of file %s\n",
         inputfile. linenum, inputfile. name ); 
   }  
   printf("\n"); 
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}


/* print s, and exit.                      */

void errorwithoutline ( s )
char *s;
{
   printf("ERROR: %s", s );
   printf("\n\n");
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}


void internalproblem( s )
char *s;
{  
   printf("\n\n");
   printf("INTERNAL PROBLEM: %s !!!", s );
   printf("\n\n");
   closefiles( );
   exit(0);
}


/******************************************************************/
/*      Functions for printing clauses:                           */
/******************************************************************/


/* print a clause in bliksem format.           */

static void printbliksemclause( f, clause )
bliksemfile *f;
int clause;
{
   int c;

   printstring( f, "(");
   printint( f, clauses [ clause + CLAUSE_NUMBER ] );
   printstring( f, ") ");

#if SHOWFILTERS
   /* Print the subsumption filters: */

   printstring( f, "[ " );
   printhex( f, clauses [ clause + CLAUSE_FILT1 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT2 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT3 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT4 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT5 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT6 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT7 ] );
   printstring( f, " ] ");
#endif

   printstring( f, "{" ); 
   printstring( f, "G" );
      printint( f, clauses [ clause + CLAUSE_GENERATION ] );
   printstring( f, ",W" );
      printint( f, clauses [ clause + CLAUSE_WEIGHT ] ); 
   printstring( f, ",D" );
      printint( f, clauses [ clause + CLAUSE_DEPTH ] ); 
   printstring( f, ",L" );
      printint( f, clauses [ clause + CLAUSE_LENGTH ] ); 
   printstring( f, ",V" );
      printint( f, clauses [ clause + CLAUSE_NRVARS ] ); 
   printstring( f, ",M" );
      printint( f, clauses [ clause + CLAUSE_END ] - 
                   clauses [ clause + CLAUSE_MAXLITS ] ); 
   printstring( f, "} " );

   c = clauses [ clause + CLAUSE_DERIVATION ];
   while( c < clauses [ clause + CLAUSE_LITERALS ] )
   { 
      switch( clauses [c] )
      {
      case RULE_RES:
         printstring( f, "R" ); 
         goto parents2; 
      case RULE_PARAMOD:
         printstring( f, "P" ); 
         goto parents2;

      case RULE_EQREFL:
         printstring( f, "Q" );
         goto parents1; 
      case RULE_EQFACT:
         printstring( f, "E" ); 
         goto parents1;
      case RULE_FACTOR:
         printstring( f, "F" );
         goto parents1;
      case RULE_RESIMP:
         printstring( f, "S" );
         goto parents1;
      case RULE_INITIAL:
         printstring( f, "I" ); 
         goto parents0;

      case SIMP_SPLITPOS:
         printstring( f, "p" );
         goto parents0;    
      case SIMP_SPLITNEG:
         printstring( f, "n" );
         goto parents0; 
      case SIMP_DEMOD:
         printstring( f, "d" );
         goto parents1;
      case SIMP_RES:
         printstring( f, "r");
         goto parents1;
      case SIMP_FACTOR:
         printstring( f, "f" );
         goto parents0;
      case SIMP_EQREFL:
         printstring( f, "q" );
         goto parents0;
      case SIMP_SELF:
         printstring( f, "s" );
         goto parents0;
      default:   
         internalproblem("Unknown rule");
         exit(0); 
      }
   
   parents0:
      c += 1;
      goto next;
   parents1:
      printstring( f, "(" );
      printint( f, clauses [ clauses[c+1] + CLAUSE_NUMBER ] );
      printstring( f, ")" );
      c += 2;
      goto next;
   parents2:
      printstring( f, "(" );
      printint( f, clauses [ clauses [c+1] + CLAUSE_NUMBER ] );
      printstring( f, "," ); 
      printint( f, clauses [ clauses [c+2] + CLAUSE_NUMBER ] );
      printstring( f, ")" );
      c += 3;
      goto next;
   next:
      if( c < clauses [ clause + CLAUSE_LITERALS ] )
         printstring( f, ";" ); 
   }

   /* the literals: */

   printstring( f, " { " ); 
   for( c = clauses [ clause + CLAUSE_LITERALS ]; 
        c < clauses [ clause + CLAUSE_END ];
        c ++ )
   {
      printterm( f, clauses [c] );
      if( c + 1 != clauses [ clause + CLAUSE_END ] )
         printstring( f, ", " );
   }
   printstring( f, " }.\n" ); 
}



/* Print a clause in prolog readable format.                     */

static void printprologclause( f, clause )
bliksemfile *f;
int clause;
{
   int lit;
   int litstart, litend;

   printstring( f, "clause( " );
   printint( f, clauses [ clause + CLAUSE_NUMBER ] );
 
   litstart =  clauses [ clause + CLAUSE_LITERALS ];
   litend = clauses [ clause + CLAUSE_END ];

   printstring( f, ", " );
   if( litstart < litend )
   { 
      printstring( f, "[ " );
      for( lit = litstart; lit < litend; lit ++ )
      {
         printterm( f, clauses [ lit ] );
         if( lit < clauses [ clause + CLAUSE_END ] - 1 )
            printstring( &outputfile, ", " );
      }
      printstring( &outputfile, " ]" );
   }
   else
      printstring( &outputfile, "[]");
   printstring( &outputfile, " )\n");
}


/* Print a clause. Choose the format.    */

void printclause( f, clause )
bliksemfile *f;
int clause;
{
   if( prologoutput )
      printprologclause( f, clause );
   else
      printbliksemclause( f, clause);
}




void printclauselist( f, clause )
bliksemfile *f;
int clause;
{
   while( clause != EMPTYCLAUSELIST )
   {
      printclause( f, clause );
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];

      if( prologoutput )
         printstring( f, ".\n" );
   }
}





/*************************************************************/
/*            Additional functions:                          */
/*************************************************************/


static void printorder( f, order )
bliksemfile *f;
int order;
{
   int ord;

   if( order < 0 )
   {
      ord = -order;
      printstring( f, "-" );
   }
   else
      ord = order;
   
   switch( ord )
   {
   case ORD_NONE:
      printstring( f, "none" );
      break;
   case ORD_DEPTH:
      printstring( f, "depth" );
      break;   
   case ORD_VARDEPTH:
      printstring( f, "vardepth" );
      break;
   case ORD_COMPL:
      printstring( f, "complexity" );
      break;
   case ORD_NRVARS:
      printstring( f, "nrvars" );
      break;
   case ORD_VD:
      printstring( f, "vd" );
      break;
   case ORD_DECISION:
      printstring( f, "varnr + vardepth" );
      break;
   case ORD_KBO:
      printstring( f, "kbo" );
      break;
   case ORD_LRPO:
      printstring( f, "lrpo" );
      break;
   case ORD_EXTTERMORDPA:
      printstring( f, "extend the termordering, first sorting on predicates" );
      break;
   case ORD_EXTTERMORDAP:
      printstring( f, "extend the termordering, first sorting on arguments" );
      break;
   case ORD_TERMORDERING:
      printstring( f, "termordering" );
      break;
   case ORD_LIFTORD:
      printstring( f, "liftord" );
      break;
   case ORD_NLO:
      printstring( f, "nlo" );
      break;
   case ORD_SPLIT:
      printstring( f, "split" );
      break;
   default: 
      internalproblem("unknown order!" );
   }
}


static void printlitselect( f, litselect )
bliksemfile *f;
int litselect;
{
   switch( litselect )
   {
   case LITSEL_NONE:
      printstring( f, "none" );
      break;
   case LITSEL_POS:
      printstring( f, "positive" );
      break;
   case LITSEL_NEG:
      printstring( f, "negative" );
      break;
   case LITSEL_POSORD:
      printstring( f, "posord" );
      break;
   case LITSEL_NEGORD:
      printstring( f, "negord" );
      break;
   default:
      internalproblem( "unknown selection function" );
   }
}


 
static void printoperatorord( f, translation )
bliksemfile *f;
int translation;
{
   int s;
   int t0, t1, i;
   int nrprinted;

   printstring( f, "\n" );
   printstring( f, "Symbols occurring in the translation:\n\n" );

   t0 = translations [ translation ]. startofresult;
   t1 = translations [ translation ]. endofresult;
   nrprinted = 0;
 
   for( s = 0; s < freeoperators; s ++ )
   {
      for( i = t0; i < t1; i ++ )
      {
         if( termspace [i] == s || s == OP_NOT )
         {
            if( nrprinted > 0 )
            {
               printstring( &outputfile, ", " );
               if( nrprinted % 1 == 0 )
                  printstring( &outputfile, "\n" );
            }
            printoperator( f, s );
            printstring( f, "  [" );
            printint( f, s );
            printstring( f, ", ");
            printint( f, arities [s] );
            printstring( f, "]      " );
            printstring( f, "(w:");
            printint( f, operatorweights [s] );
            printstring( f, ", o:");
            printint( f, operatororder [s] );
            printstring( f, ", a:");
            printint( f, operatorassoc [s] );
            printstring( f, ", s:");
            printint( f, operatorselect [s] ); 
            printstring( f, ", b:");
            printint( f, operatornoparamod [s] );
            printstring( f, ")" );

            nrprinted ++ ;
            goto printed;
         }
      }
   printed:
      ;
   }
   printstring( f, ".\n\n" );
}


static void printsubstype( f, substype )
bliksemfile *f;
int substype;
{
   switch( substype )
   {
   case SUBS_STANDARD:
      printstring( f, "standard" );
      break;
   case SUBS_SNL1:
      printstring( f, "snl1" );
      break;
   case SUBS_EQREWR:
      printstring( f, "eqrewr" );
      break;
   default:
      internalproblem("Unknown substype!");
   }
}


static void printstrategy( f, strategy )
bliksemfile *f;
int strategy;
{
   switch( strategy )
   {
   case STRAT_AUTOMODE:
      printstring( f, "automode" );
      break;
   case STRAT_AORDERED:
      printstring( f, "A-ordered resolution" );
      break;
   case STRAT_NONLIFT:
      printstring( f, "resolution with non-liftable order" );
      break;
   case STRAT_HYPER:
      printstring( f, "positive hyperresolution" );
      break;
   case STRAT_SUPERPOS:
      printstring( f, "superposition without selection" );
      break; 
   case STRAT_SUPERPOSSEL:
      printstring( f, "superposition with negative selection" ); 
      break;  
   case STRAT_EPLUS:
      printstring( f, "procedure for E-plus class" );
      break; 
   default:
      internalproblem( "printing unknown strategy" );
   }
}
 
 
static void printoptions( f, conversion, strategy )
bliksemfile *f;
int conversion;
int strategy;
{
   int i;

   printstring( f, "\n\n" );
   printstring( f, "Options Used:\n\n" ); 

   printstring( f, "conversion = " );
   printfullconversion( f, conversion );
   printstring( f, "\n" );

   printstring( f, "strategy = " );
   printstrategy( f, strategy );
   printstring( f, "\n" );  

   printstring( f ,"\n" ); 

   printstring( f, "useres =            " );
   printint( f, useres );
   printstring( f, "\n" );

   printstring( f, "useparamod =        " );
   printint( f, useparamod );
   printstring( f, "\n" );

   printstring( f, "useeqrefl =         " );
   printint( f, useeqrefl );
   printstring( f, "\n" );

   printstring( f, "useeqfact =         " );
   printint( f, useeqfact );
   printstring( f, "\n" );

   printstring( f, "usefactor =         " );
   printint( f, usefactor );
   printstring( f, "\n" );

   printstring( f, "usesimpdemod =      " );
   printint( f, usesimpdemod ); 
   printstring( f, "\n" ); 

   printstring( f, "usesimpres =        " );
   printint( f, usesimpres );
   printstring( f, "\n" ); 

   printstring( f, "\n" );

   printstring( f, "resimpinuse      =  " );
   printint( f, resimpinuse );
   printstring( f, "\n" );

   printstring( f, "resimpclauses =     " );
   printint( f, resimpclauses );
   printstring( f, "\n" );

   printstring( f, "substype =          " );
   printsubstype( f, substype );
   printstring( f, "\n" );

   printstring( f, "backwardsubs =      " );
   printint( f, backwardsubs );
   printstring( f, "\n" );

   printstring( f, "selectoldest =      " );
   printint( f, selectoldest );
   printstring( f, "\n" );

   if( nrlitorderings > 0 )
      printstring( f, "\n" ); 

   for( i = 0; i < nrlitorderings; i ++ )
   {
      printstring( f, "litorderings [");
      printint( f, i );
      printstring( f, "] =  " );
      printorder( f, litorderings [i] );
      printstring( f, "\n");

   }
   printstring( f, "\n" );

   printstring( f, "termordering =      " );
   printorder( f, termordering );
   printstring( f, "\n" );
   printstring( f, "\n" );

   printstring( f, "litapriori =        " );
   printint( f, litapriori );
   printstring( f, "\n" );

   printstring( f, "termapriori =       " );
   printint( f, termapriori );
   printstring( f, "\n" );

   printstring( f, "litaposteriori =    " );
   printint( f, litaposteriori );
   printstring( f, "\n" );

   printstring( f, "termaposteriori =   " );
   printint(f, termaposteriori );
   printstring( f, "\n" );

   printstring( f, "demodaposteriori =  " );
   printint( f, demodaposteriori );
   printstring( f, "\n" );

   printstring( f, "ordereqreflfact =   " );
   printint( f, ordereqreflfact );
   printstring( f, "\n\n" );

   printstring( f, "litselect =         " );
   printlitselect( f, litselect );
   printstring( f, "\n\n" );

   printstring( f, "maxweight =         " );
   printint( f, maxweight );
   printstring( f, "\n" );

   printstring( f, "maxdepth =          " );
   printint( f, maxdepth );
   printstring( f, "\n" );

   printstring( f, "maxlength =         " );
   printint( f, maxlength );
   printstring( f, "\n" );

   printstring( f, "maxnrvars =         " );
   printint( f, maxnrvars );
   printstring( f, "\n" );

   printstring( f, "excuselevel =       " );
   printint( f, excuselevel );
   printstring( f, "\n" );

   printstring( f, "increasemaxweight = " );
   printint( f, increasemaxweight );
   printstring( f, "\n\n" );

   printstring( f, "maxselected =       " );
   printint( f, maxselected );
   printstring( f, "\n" );

   printstring( f, "maxnrclauses =      " );
   printint( f, maxnrclauses );
   printstring( f, "\n\n" );
   
   printstring( f, "showgenerated =    " );
   printint( f, showgenerated );
   printstring( f, "\n" );

   printstring( f, "showkept =         " );
   printint( f, showkept );
   printstring( f, "\n" );

   printstring( f, "showselected =     " );
   printint( f, showselected );
   printstring( f, "\n" );

   printstring( f, "showdeleted =      " );
   printint( f, showdeleted );
   printstring( f, "\n" );

   printstring( f, "showresimp =       " );
   printint( f, showresimp );
   printstring( f, "\n" );

   printstring( f, "showstatus =       " );
   printint( f, showstatus );
   printstring( f, "\n");
   printstring( f, "\n" );

   printstring( f, "prologoutput =     " );
   printint( f, prologoutput );
   printstring( f, "\n" );

   printstring( f, "nrgoals =          " );
   printint( f, nrgoals );
   printstring( f, "\n" );

   printstring( f, "totalproof =       " );
   printint( f, totalproof );
   printstring( f, "\n" );

   printstring( f, "refuseequality =   " );
   printint( f, refuse_equality );
   printstring( f, "\n" );

   if( nrtranslations != 1 )
      internalproblem( "nrtranslations" ); 
   printoperatorord( f, 0 );

}




/***************************************************************/
/*   Functions belonging to the automode.                      */
/***************************************************************/

/* Give the percentage of good Horn clauses.   */

static double pcthorn( clause )
int clause;
{
   int total, horn;
   int nrpos;
   int cl, lit;

   total = 0;
   horn = 0;
   while( clause != EMPTYCLAUSELIST ) 
   {
      nrpos = 0;
      for( cl = clauses [ clause + CLAUSE_LITERALS ];
           cl < clauses [ clause + CLAUSE_END ];
           cl ++ )
      {
         lit = clauses [ cl ];
         if( termspace [ lit ] != OP_NOT )
            nrpos ++ ;
      }
      if( nrpos <= 1 )
         horn ++ ;
      total ++ ; 

      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
   if( total != 0 )
      return ( (double) horn ) / ( (double) total ); 
   else
      return 1.0;
}


/* Produce pct. of equality literals in list.     */

static double pctequality( list )
int list;
{
   int nrliterals;
   int nrequality;
   int cl, at;

   nrliterals = 0;
   nrequality = 0;
   while( list != EMPTYCLAUSELIST ) 
   {
      for( cl = clauses [ list + CLAUSE_LITERALS ];
           cl < clauses [ list + CLAUSE_END ];
           cl ++ )
      {
         at = clauses [ cl ];
         if( termspace [ at ] == OP_NOT )
            at ++ ;
         if( termspace [ at ] == OP_EQUAL || termspace [ at ] == OP_REWRITES )
            nrequality ++ ;
         nrliterals ++;
      }
      list = clauses [ list + CLAUSE_NEXTINLIST ];
   }
   if( nrliterals > 0 )
      return (double) nrequality / (double) nrliterals;
   else
      return 0.0; 
}
 




/* Give all symbols equal weight, with the exception of not and equality.  */

static void setweights( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
   {
      if( operators [i]. flags & FLAG_SKOLEM )
      {
         operatorweights [i] = 1;
      }
      else
      { 
         operatorweights [i] = 1;
      }
   }
   operatorweights [ OP_NOT ] = 0;   
   operatorweights [ OP_EQUAL ] = 0;
   operatorweights [ OP_REWRITES ] = 0;
}


/* Make all symbols in selectable by the selection function.   */

static void setselects( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
      operatorselect [i] = 1;   
}


static void setassocs( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
      operatorassoc [i] = 1;

   operatorassoc [ OP_EQUAL ] = 0;
   operatorassoc [ OP_REWRITES ] = 0; 
}


static void setnoparamods( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
   {
      if( ( operators [i]. flags & FLAG_SUBFORM ) ||
          ( operators [i]. flags & FLAG_SKOLEM  ))
      {
         operatornoparamod [i] = 1;   
      } 
      else
      {
         operatornoparamod [i] = 0;
      } 
   }
}

/****************************************************************/
/* Function shouldbesmaller( op1, op2 ) returns 1 if op1        */
/* should be smaller in the precedence than op2.                */
/****************************************************************/


static int shouldbesmaller_skolem( op1, op2 )
int op1, op2; 
{
   if( ! ( operators [ op1 ]. flags & FLAG_SKOLEM ))
   {
      if( ( operators [ op2 ]. flags & FLAG_SKOLEM ))
         return 1;
   }
   return 0; 
}


/* Symbols introduced in the structural clause transformation */
/* are put smaller than other symbols.                        */

static int shouldbesmaller_struct( op1, op2 )
int op1, op2;
{
   if( operators [ op1 ]. flags & FLAG_SUBFORM ) 
   {
      if( ! ( operators [ op2 ]. flags & FLAG_SUBFORM ))
      {
         return 1;
      }
   }
   return 0;
}


 
static int shouldbesmaller_arity( op1, op2 )
int op1, op2;
{
   if( arities [ op1 ] < arities [ op2 ] )
      return 1;
   return 0;
}


static int shouldbesmaller( op1, op2 )
int op1, op2;
{
   if( shouldbesmaller_struct( op1, op2 ))
      return 1;
   if( shouldbesmaller_struct( op2, op1 ))
      return 0; 

   if( shouldbesmaller_skolem( op1, op2 )) 
      return 1;
   if( shouldbesmaller_skolem( op2, op1 )) 
      return 0;

   if( shouldbesmaller_arity( op1, op2 ))
      return 1;
   if( shouldbesmaller_arity( op2, op1 ))
      return 0; 

   /* We want the order to be total, so we complete it by alfabetic */
   /* sorting.                                                      */

   if( op1 < OP_NONPRIMITIVE && op2 >= OP_NONPRIMITIVE ) 
      return 1;
   if( op2 < OP_NONPRIMITIVE )
      return 0; 

   if( strcmp( operatorstring( op1 ), operatorstring( op2 )))
   {
      return 1;
   } 
   return 0;
}


/* Produce an order on the symbols, using shouldbesmaller.          */

static void setordering( )
{
   int i;
   int smallest;
   int counter;

   for( i = 0; i < freeoperators; i ++ )
      operatororder [i] = -1;

   operatororder [ OP_EQUAL ] = 0;
   operatororder [ OP_REWRITES ] = 0;
   counter = 1;

   while( 1 )
   {
      /* Find the smallest operator, which still has a mark -1. */

      smallest = -1;
      for( i = 0; i < freeoperators && smallest == -1; i ++ )
      {
         if( operatororder [i] == -1 )
            smallest = i;           
      }

      /* If no marks -1 left, then we are ready.    */

      if( smallest == -1 )
         return;

      for( i = smallest; i < freeoperators; i ++ )
      {      
         if( operatororder [i] == -1 && shouldbesmaller( i, smallest ))
         {
            smallest = i;
         }
      }

      operatororder [ smallest ] = counter;
      counter ++ ; 
   }
}


/* Set the parameters of the symbols.     */ 

static void setwoasp( )
{
   setselects( );
   setweights( );
   setassocs( ); 
   setnoparamods( ); 
   setordering( ); 
}



/*******************************************************/
/* Option modifiers.                                   */
/*******************************************************/


static void mod_increasemaxweight( )
{
   maxweight = 10;
   increasemaxweight = 1;
   excuselevel = 3; 
}


static void mod_aposteriori( )
{
   if( litapriori )
   {
      litaposteriori = 1;
   }

   if( termapriori )
   {
      termaposteriori = 1;
      demodaposteriori = 1;
   } 
}


/********************************************************/
/* Options setting functions:                           */
/********************************************************/


static void strategy_aordered( )
{
   useres = 1;
   useparamod = 0;
   
   useeqrefl = 0;
   useeqfact = 0;

   usefactor = 1;

   usesimpdemod = 0;
   usesimpres = 3;

   substype = SUBS_STANDARD;

   litorderings [0] = ORD_SPLIT;
   litorderings [1] = ORD_LIFTORD;
   nrlitorderings = 2;

   litapriori = 1;
   termapriori = 0;
   litaposteriori = 0;
   termaposteriori = 0;
   demodaposteriori = 0;

   setwoasp();
}


static void strategy_nonlift( )
{
   internalproblem( "nonlift" );
}


static void strategy_superpos( )
{
   useres = 1;
   useparamod = 1;

   useeqrefl = 1;
   useeqfact = 1;
   usefactor = 1;

   usesimpdemod = 4;
   usesimpres = 4;

   substype = SUBS_EQREWR;
   backwardsubs = 1;

   litorderings [0] = ORD_EXTTERMORDAP;
   nrlitorderings = 1;

   termordering = ORD_KBO;

   litapriori = 1; 
   termapriori = 1; 

   litselect = LITSEL_NONE;

   setwoasp( ); 
}


static void strategy_superpossel( )
{
   useres = 1;
   useparamod = 1;
   useeqrefl = 1;
   useeqfact = 1;
   usefactor = 1;

   usesimpdemod = 4;
   usesimpres = 4;

   substype = SUBS_EQREWR;
   backwardsubs = 1;

   nrlitorderings = 1;
   litorderings [0] = ORD_EXTTERMORDAP;

   termordering = ORD_KBO;

   litapriori = 1;
   termapriori = 1;

   litselect = LITSEL_NEG;
  
   setwoasp( );

   operatorweights [ OP_NOT ] = 1;
}


/* strategy_hyper is complete only when there is no equality. */

static void strategy_hyper( )
{
   useres = 1;
   useparamod = 0;
   useeqrefl = 0;
   useeqfact = 0;
   usefactor = 1;

   usesimpdemod = 0;
   usesimpres = 3;
   
   nrlitorderings = 1;
   litorderings [0] = ORD_LIFTORD;

   termordering = ORD_NONE;

   litapriori = 1;
   termapriori = 0;
   litaposteriori = 0;
   termaposteriori = 0;
   demodaposteriori = 0;

   litselect = LITSEL_NEG;

   setwoasp( ); 

   operatorweights [ OP_NOT ] = 1;
}



/* Make strategy selection. Return the conversion that should be used. */

static int automode( conv )
int conv;
{

   int i, strat;
   int freetermspace2, freeclauses2;
   double eq, horn;
   int strategy;

   printstring( &outputfile, "Automatic Strategy Selection\n\n" );

   freetermspace2 = freetermspace;
   freeclauses2 = freeclauses;

   i = mainloop( conv, 1 );

   eq = pctequality( listclauses );
   horn = pcthorn( listclauses );

#if 1
   printf("percentage equality = %lf, percentage horn = %lf\n", eq, horn );
#endif

   if( eq > 0.0 )
   {
      if( horn >= 0.8 )
      {
         strategy = STRAT_SUPERPOSSEL;
      }
      else
      {
         strategy = STRAT_SUPERPOS;
      }
   }
   else
   {
      if( horn >= 0.8 )
      {
         strategy = STRAT_HYPER;      
      }
      else
      {
         strategy = STRAT_AORDERED;
      }
   }

   freetermspace = freetermspace2;
   freeclauses = freeclauses2; 
   return strategy;
}



int readoptions( conversion, strategy, argc, argv )
int *conversion;
int *strategy;
int argc;
char *argv [];
{
   int i;

   i = 1;
   
readoption:
   if( i + 2 < argc && strcmp( argv [i], "-c" ) == 0 ) 
   {
      if( strcmp( argv [i+1], "none" ) == 0 )
      {
         *conversion = FULLCONV_NONE;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv [i+1], "standard" ) == 0 )
      {
         *conversion = FULLCONV_STANDARD;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv [i+1], "guarded" ) == 0 )
      {
         *conversion = FULLCONV_GUARDED;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv [i+1], "struct" ) == 0 )
      {
         *conversion = FULLCONV_STRUCT;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv [i+1], "structos" ) == 0 )
      {
         *conversion = FULLCONV_STRUCTOS;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv[i+1], "2var" ) == 0 )
      {
         *conversion = FULLCONV_2VAR;
         i += 2;
         goto readoption;
      } 
      printf( "ERROR: Did not recognize conversion %s", argv [i+1] );  
      printf("\n\n");
      exit(0);
   } 

   if( i + 2 < argc && strcmp( argv [i], "-s" ) == 0 )
   {
      if( strcmp( argv [i+1], "automode" ) == 0 )
      {
         *strategy = STRAT_AUTOMODE;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv [i+1], "aordered" ) == 0 )
      {
         *strategy = STRAT_AORDERED; 
         i += 2;
         goto readoption;
      }
      if( strcmp( argv[i+1], "nonlift" ) == 0 )
      {
         *strategy = STRAT_NONLIFT;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv[i+1], "hyper" ) == 0 )
      {
         *strategy = STRAT_HYPER;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv[i+1], "superpos" ) == 0 )
      {
         *strategy = STRAT_SUPERPOS;
         i += 2;
         goto readoption;
      }
      if( strcmp( argv[i+1], "superpossel" ) == 0 )
      {
         *strategy = STRAT_SUPERPOSSEL;
         i += 2;
         goto readoption;
      }
      printf( "ERROR: Did not recognize strategy %s", argv [i+1] );
      printf("\n\n");
      exit(0);
   }

   if( i + 1 < argc && strcmp( argv [i], "-neq" ) == 0 )
   {
      refuse_equality = 1;
      i += 1;
      goto readoption;
   }
   return i;
}






/************************************************************/
/*                       main                               */
/************************************************************/



int main( argc, argv )
int argc;
char *argv [];
{
   int i; 
   int freetermspace2, freeclauses2; 
   int conv;

   int conversion;
   int strategy;

   if( sizeof(int) < 4 )
   {
      printf("\n"); 
      printf("ERROR: need at least 32 bit integers!\n\n");
      return 0;
   }

   init_termspace( ); 
   init_clauses( );
   init_justifications( );

   conversion = FULLCONV_STRUCT; 
   strategy = STRAT_AUTOMODE;
   
   i = readoptions( &conversion, &strategy, argc, argv );

   initinoutput( i, argc, argv );

   (void) parser ( );

   printstring( &outputfile, "Bliksem 1.12 A\n\n" ); 
   printstring( &outputfile,"\n"); 

   maketermends( 0, freetermspace ); 
   
   /* Default settings of flags:   */

   useres = 0;
   useparamod = 0;
   useeqrefl = 0;
   useeqfact = 0;
   usefactor = 0;

   usesimpdemod = 0;
   usesimpres = 0;

   resimpinuse = 1000;
   resimpclauses = 20000; 

   substype = SUBS_STANDARD; 
   backwardsubs = 1; 

   selectoldest = 5;

   nrlitorderings = 0;
   termordering = ORD_NONE;

   litapriori = 0;
   termapriori = 0;
   litaposteriori = 0;
   termaposteriori = 0;
   demodaposteriori = 0;
   ordereqreflfact = 0;

   litselect = LITSEL_NONE; 

   maxweight =    32000;
   maxdepth =     32000;
   maxlength =    MAXCLAUSELEN - 5;
   maxnrvars =    MAXNRVARS - 5; 

   excuselevel = 1;
   increasemaxweight = 0;

   maxselected =  10000000;
   maxnrclauses = 10000000;

   showgenerated = 0;
   showkept = 0;
   showselected = 0;
   showdeleted = 0;
   showresimp = 1;
   showstatus = 2000;

   prologoutput = 0;
   nrgoals = 10000000;
   totalproof = 0;


   startofinput = 0;
   endofinput = freetermspace; 


   /* Construct one translation.   */ 

   nrtranslations = 0;
   conv = makeconversion( startofinput, endofinput, conversion );
   if( nrtranslations != 1 )
   {
      internalproblem( "nrtranslations" );
   }
   printstring( &outputfile, "\n" );

   if( strategy == STRAT_AUTOMODE )
   {
      strategy = automode( conv );
   }

   switch( strategy )
   {
   case STRAT_AORDERED:
      strategy_aordered( );
      break;
   case STRAT_NONLIFT:
      strategy_nonlift( );
      break;
   case STRAT_HYPER:
      strategy_hyper( );
      break; 
   case STRAT_SUPERPOS:
      strategy_superpos( );  
      break; 
   case STRAT_SUPERPOSSEL:
      strategy_superpossel( );
      break; 
   default:
      internalproblem( "unknown strategy" );
   }

#if 0 
   mod_aposteriori( );
#endif

#if 1 
   mod_increasemaxweight( ); 
#endif

   freetermspace2 = freetermspace;
   freeclauses2 = freeclauses;

   printoptions( &outputfile, conversion, strategy );

   if( refuse_equality ) 
   {
      int t0, t1; 

      t0 = translations [ conv ]. startofresult;
      t1 = translations [ conv ]. endofresult; 

      for( i = t0; i < t1; i ++ )
      {
         if( termspace [i] == OP_EQUAL || termspace [i] == OP_REWRITES )
         {
            printstring( &outputfile, "abandoned because of equality\n");
            printstring( &outputfile, "\n" );
            return 0;
         }
      }
   }

#if 0
   /* Always quit.    */

   printstring( &outputfile, "no proof search\n\n" );
   return 0; 
#endif 

restart:
   printstring( &outputfile, "\nStarting Search:\n\n" ); 

   i = mainloop( conv, 0 );

   if( ! i && increasemaxweight )
   {
      freetermspace = freetermspace2;
      freeclauses = freeclauses2;   

      /* maxnrclauses *= 2;                 */
      maxweight += increasemaxweight ; 
      printstring( &outputfile, "Increased maxweight to ");
      printint( &outputfile, maxweight );
      printstring( &outputfile, "\n" );

      goto restart;
   }
   printstatistics( &outputfile ); 
   return 0;
}





