

/*********************************************************/
/* File proofcheck.c contains functions for checking the */
/* proof and printing it in full detail.                 */
/* The file is self-supporting because I want to keep    */
/* it away from the rest of the theorem prover.          */
/* Hans de Nivelle, July 1999.                           */
/*********************************************************/

/* We currently have a bug that sometimes the substitution  */
/* includes variables that do not occur in the clause.      */


#define CLASS extern
#include "basics.h"
#include "extfunc.h"
#include "resolution.h"

#define MAXPARAMOD   10
   /* Maximal number of replacements in paramodulation step.          */
   /* This is necessary because Bliksem replaces all occurrences in a */
   /* paramodulation step.                                            */

#define MAXWEIGHT    100 
   /* Maximal weight of a clause.                                     */

/************************************************************/
/* Storage conventions:                                     */
/* An END OF A RECORD is stored as an index in              */
/* justifications [].                                       */
/* A CLAUSE is stored as an index in clauses.               */
/* A LITERAL in a clause is stored as an index in clauses,  */
/* so clauses [ LITERAL ] equals the literal.               */
/* SUBSTITUTIONS are stored mixed, and unextended,          */
/* as quadruples variable, varorigin, term, termorigin.     */
/* Substitutions start with an end reference.               */
/* RENAMINGS are stored as substitutions, with the origin   */
/* fields 0, (or don't care)                                */
/* TERMS are stored in prefix notation, without termends    */
/* references.                                              */
/************************************************************/


/************************************************************/
/* Used for storing the derivation. Everything is stored.   */
/************************************************************/



#define JUST_FACTOR             300
   /* Followed by:                                          */
   /* End of record.                                        */
   /* Derived clause.                                       */
   /* Parent clause.                                        */
   /* Literal factored upon in parent clause.               */
   /* Other literal factored upon in parent clause.         */
   /*      (the one that is deleted)                        */
   /* Substitution used                                     */
   /* Renaming used.                                        */
   
   /* Note that we see clauses as lists. Because of         */
   /* this it is possible that the substitution is empty.   */

#define JUST_EQFACT             301
   /* Followed by:                                          */
   /* End of record.                                        */
   /* Derived clause.                                       */
   /* Parent clause.                                        */
   /* Literal t1 = t2.                                      */
   /* Literal t1 = t3.                                      */
   /*      (The result is t2 != t3, t1 = t3 )               */
   /* Substitution used                                     */
   /* Renaming used.                                        */
   

#define JUST_PARAMOD            302
   /* Followed by:                               */
   /* End of record.                             */
   /* Derived clause.                            */
   /* From parent clause.                        */
   /* From literal paramodulated upon.           */
   /* Into parent clause.                        */
   /* Into literal paramodulated upon.           */
   /* Into term paramodulated upon. Stored as    */
   /*    index in the term after the             */
   /*    substitution.                           */  
   /* Mention the whole path, rather than this   */
   /* index.                                     */
   /* Substitution used                          */
   /* Renaming used.                             */

#define JUST_RESO               303
   /* Followed by:                                       */
   /* End of record.                                     */
   /* Derived clause.                                    */
   /* Parent clause containing negative literal.         */
   /* Negative literal resolved upon.                    */
   /* Parent clause containing positive literal.         */
   /* Positive literal resolved upon.                    */
   /* Substitution used                                  */
   /* Renaming used.                                     */

#define JUST_SUBSUMES          304
   /* Followed by:                      */
   /* End of record.                    */
   /* Derived clause.                   */
   /* Parent clause.                    */
   /* Substitution used.                */


   /* Would it be a good idea to mention not only the permutation, */
   /* but also the weakening?                                      */


#define JUST_EQSWAP             305
   /* Followed by:                             */
   /* End of record.                           */
   /* Derived clause.                          */
   /* Parent clause.                           */
   /* Literal in which equality is swapped.    */
   /* Substitution used.                       */
   /* Renaming used.                           */

#define JUST_EQREFL             306
   /* Followed by:                             */
   /* End of record.                           */
   /* Derived clause.                          */
   /* Parent clause.                           */
   /* Literal ! ( t1 = t2 ) that is deleted.   */
   /* Substitution used.                       */
   /* Renaming used.                           */



/***********************************************************/
/* Here the proof for each proven goal will be collected.  */
/***********************************************************/


#define MAXPROOF 5000
static int proof [ MAXPROOF ];
static int freeproof;


/***************************************************/
/*      Used for the rededuction process:          */
/***************************************************/

static int just_initials1, just_initials2; 
   /* clauses [ just_initials1 ... just_initials2 ] contains the */
   /* initial clauses.                                           */


static int just_goal;
   /* Goal clause.                          */
static int just_parent1, just_parent2;
   /* The parents of the main derivation.   */

static int just_nrkept;  
   /* Used to be able to give a distinct number to each clause.  */
   /* The first numbers are reserved for the (real) initial      */
   /* clauses.                                                   */ 

static int just_maxweight;
   /* Larger clauses are thrown away.                            */

static int just_justificationsbase;
   /* There may be justifications produced by the conversion.    */


static jmp_buf just_success;    /* Jumped to on succes.          */



/**********************************************************************/
/* Specialized matching and unification functions that do not         */
/* distinguish ==> from =. They are considered the same symbol in     */
/* this file, but NOT commutative.                                    */
/* They are put here because at all other points commutativity is     */
/* included. This however contradicts the nature of the proofs that   */
/* we want to construct.                                              */
/**********************************************************************/


/* Match lit1 into lit2, ignoring the difference between = and ==>,   */
/* but NOT using commutativity.                                       */
 
static int matcheqrewr( lit1, lit2 )
int lit1;
int lit2;
{
   int at1, at2;
 
   if( termspace [ lit1 ] == OP_NOT )
   {
      if( termspace [ lit2 ] == OP_NOT )
      {
         at1 = lit1 + 1;
         at2 = lit2 + 1;
      }
      else
         return 0;
   }
   else
   {
      if( termspace [ lit2 ] == OP_NOT )
         return 0;
      else
      {
         at1 = lit1;
         at2 = lit2;
      }
   }
 
   if( ( termspace [ at1 ] == OP_EQUAL || termspace [ at1 ] == OP_REWRITES ) 
         &&
       ( termspace [ at2 ] == OP_EQUAL || termspace [ at2 ] == OP_REWRITES ))
   {
      if( match( at1 + 1, at2 + 1 ) &&
          match( termends [ at1 + 1 ], termends [ at2 + 1 ] ))
         return 1;
      else
         return 0;
   }
   else
   {
      return match( at1, at2 );
   }
}
 

/* unify lit1 with lit2 ignoring the difference between = and ==>.   */
/* However do NOT unify t1 = t2 with t2 = t1, or t1 ==> t2 with      */
/* t2 ==> t1.                                                        */

static int unifyeqrewr( lit1, orig1, lit2, orig2 )
int lit1;
int orig1;
int lit2;
int orig2;
{
   int at1, at2;

   if( termspace [ lit1 ] == OP_NOT )
   {
      if( termspace [ lit2 ] == OP_NOT )
      {
         at1 = lit1 + 1;
         at2 = lit2 + 1;
      }
      else
         return 0;
   }
   else
   {
      if( termspace [ lit2 ] == OP_NOT )
         return 0;
      else
      {
         at1 = lit1;
         at2 = lit2;
      }
   }
 
   if( ( termspace [ at1 ] == OP_EQUAL || termspace [ at1 ] == OP_REWRITES ) 
         &&
       ( termspace [ at2 ] == OP_EQUAL || termspace [ at2 ] == OP_REWRITES ))
   {
      if( unify( at1 + 1, orig1, at2 + 1, orig2 ) &&
          unify( termends [ at1 + 1 ], orig1, termends [ at2 + 1 ], orig2 ))
         return 1;
      else
         return 0;
   }
   else
   {
      return unify( at1, orig1, at2, orig2 );
   }
}


/* return 1 if term1 and term2 are equal, ignoring the difference */
/* between = and ==>,  but NOT using commutativity.               */

static int termequaleqrewr( term1, term2 )
int term1;
int term2;
{
   int endterm1;
   int a;

   endterm1 = termends [ term1 ];
   if( termspace [ term1 ] == OP_NOT && 
       termspace [ term2 ] == OP_NOT ) 
   {
      term1 ++ ;
      term2 ++ ;
   }
   if( ( termspace [ term1 ] == OP_EQUAL ||
         termspace [ term1 ] == OP_REWRITES ) &&
       ( termspace [ term2 ] == OP_EQUAL ||
         termspace [ term2 ] == OP_REWRITES ) ) 
   {
      term1 ++ ; 
      term2 ++ ;
   }

   while( term1 < endterm1 )
   {
      a = termspace [ term1 ++ ];
      if( a != termspace [ term2 ++ ] )
         return 0;
   }
   return 1;
}



/***********************************************************/
/*       Store report in justifications [].                */
/***********************************************************/



/* Store a substitution in justifications [].  */

static void storesubstitution( )
{
   int i;
   int substbase;

   substbase = freejustifications; 
   freejustifications ++ ; 
 
   for( i = 0; i < freesubst; i ++ ) 
   {
      if( freejustifications >= maxjustifications10 ) 
         more_justifications( );

      justifications [ freejustifications ++ ] = subst [i]. var;
      justifications [ freejustifications ++ ] = subst [i]. varorig;
      justifications [ freejustifications ++ ] = subst [i]. term; 
      justifications [ freejustifications ++ ] = subst [i]. termorig; 
   }
   justifications [ substbase ] = freejustifications; 
}


/* Store a matching as a substitution in justifications []. */

static void storematching( )
{
   int i;
   int substbase;

   substbase = freejustifications;
   freejustifications ++ ;

   for( i = 0; i < freematching; i ++ ) 
   {
      if( freejustifications >= maxjustifications10 ) 
         more_justifications( ); 
      
      justifications [ freejustifications ++ ] = matching [i]. var;
      justifications [ freejustifications ++ ] = 0;
      justifications [ freejustifications ++ ] = matching [i]. term;
      justifications [ freejustifications ++ ] = 1;
   }
   justifications [ substbase ] = freejustifications;
}


/* Store a renaming as a substitution in justifications []. */

static void storerenaming( )
{
   int i;
   int substbase;

   substbase = freejustifications;
   freejustifications ++ ; 

   for( i = 0; i < freerenaming; i ++ ) 
   {
      if( freejustifications >= maxjustifications10 ) 
         more_justifications( );

      justifications [ freejustifications ++ ] = renaming [i]. var;
      justifications [ freejustifications ++ ] = renaming [i]. varorig;  
      justifications [ freejustifications ++ ] = - ( i + 1 );
      justifications [ freejustifications ++ ] = 7;
   }
   justifications [ substbase ] = freejustifications; 
}


/********************************************************************/
/*                   printjustification.                            */
/********************************************************************/


static void readsubstitution( just )
int *just;
{
   int i;

   i = *just + 1;
   freesubst = 0;
   while( i < justifications [ *just ] )
   {
      subst [ freesubst ]. var = justifications [ i ++ ];
      subst [ freesubst ]. varorig = justifications [ i ++ ];
      subst [ freesubst ]. term = justifications [ i ++ ];
      subst [ freesubst ]. termorig = justifications [ i ++ ] ;
      freesubst ++ ; 
   }
   *just = i;
}


static void readrenaming( just )
int *just;
{
   int i;

   i = *just + 1;
   freerenaming = 0;
   while( i < justifications [ *just ] )
   {
      renaming [ freerenaming ]. var = justifications [ i ++ ] ;
      renaming [ freerenaming ]. varorig = justifications [ i ++ ]; 

      if( justifications [ i ++ ] != - ( freerenaming  + 1 ) ) 
         internalproblem( "problem in readrenaming" );
      if( justifications [ i ++ ] != 7 ) 
         internalproblem( "problem with readrenaming" ); 

      freerenaming ++ ; 
   }
   *just = i;
}


static void just_printsubst( f, orig )
bliksemfile *f;
int orig;
{
   int i;

   int minvar;
   int var, res;
   int termspacebase; 

   minvar = 0;
   for( i = 0; i < freesubst; i ++ )
   {
      if( subst[i]. var < minvar && subst [i]. varorig == orig )
         minvar = subst [i]. var;
   }

   for( i = 0; i < freerenaming; i ++ )
   {
      if( renaming [i]. var < minvar && renaming [i]. varorig == orig )
         minvar = renaming [i]. var;       
   }

   minvar -- ;

   if( prologoutput )
   {
      printstring( f, "substitution( " );
      printint( f, orig );
      printstring( f, ", [" );
   }
   else
   {
      printstring( f, "substitution");
      printint( f, orig );
      printstring( f, ":\n" ); 
   }

   for( i = -1; i > minvar; i -- )
   {
      if( freetermspace >= maxtermspace10 )
         more_termspace( ); 

      termspacebase = freetermspace;
      var = freetermspace;
      termspace [ freetermspace ++ ] = i;
      maketermends( var, freetermspace );

      res = ressubst( var, orig );
      maketermends( res, freetermspace );

      if( prologoutput )
      {
         if( i == -1 )
            printstring( f, " " );
         else
            printstring( f, ", " );
         printstring( f, ":=( " );
         printterm( f, var );
         printstring( f, ", " );
         printterm( f, res );
         printstring( f, " )" );
      }
      else
      {
         printstring( f, "   ");  
         printterm( f, var );
         printstring( f, " := " );
         printterm( f, res );
         printstring( f, "\n" ); 
      }
      freetermspace = termspacebase;
   }
   if( prologoutput )
      printstringnnl( f, "] )" );
   else
      printstring( f, "end\n" );
}
 
/* print the permutation with which clause1 subsumes clause2. */

static void just_printpermutation( f, clause1, clause2 )
bliksemfile *f;
int clause1;
int clause2;
{
   int i, j;
   int res, termspacebase;

   if( prologoutput )
      printstring( f, "permutation( 0, [" );
   else
      printstring( f, "permutation0:\n" );

   for( i = clauses [ clause1 + CLAUSE_LITERALS ]; 
        i < clauses [ clause1 + CLAUSE_END ]; 
        i ++ )
   {
      if( prologoutput )
      {
         if( i == clauses [ clause1 + CLAUSE_LITERALS ] )
            printstring( f, " " );
         else
            printstring( f, ", " );
         printstring( f, "==>( " );
         printint( f, i - clauses [ clause1 + CLAUSE_LITERALS ] );
         printstring( f, ", " );
      }
      else
      {
         printstring( f, "   " );
         printint( f, i - clauses [ clause1 + CLAUSE_LITERALS ] );
         printstring( f, " ==> " );
      }
 
      termspacebase = freetermspace;
      res = ressubst( clauses [i], 0 );
      maketermends( res, freetermspace );

      for( j = clauses [ clause2 + CLAUSE_LITERALS ];
           j < clauses [ clause2 + CLAUSE_END ];
           j ++ )
      {
         if( termequaleqrewr( res, clauses [j] ))
         {
            printint( f, j - clauses [ clause2 + CLAUSE_LITERALS ] );
            goto found;
         }
      }
      internalproblem( "problem in printpermutation" ); 
   found:
      freetermspace = termspacebase; 
      if( prologoutput )
         printstring( f, " )" );
      else 
         printstring( f, "\n" );
   }
   if( prologoutput )
      printstringnnl( f, "] )" );
   else
      printstring( f, "end\n" );
}



static void printbinaryjust( f, s, result, clause1, cl1, lit1, 
                                           clause2, cl2, lit2 )
bliksemfile *f;
char s [];
int result;
int clause1, cl1, lit1;
int clause2, cl2, lit2;
{

   if( prologoutput )
   {
      printstring( f, s );
      printstringnnl( f, "(\n");
      printclause( f, result );
      printstring( f, ", " );

      printclause( f, clause1 );  printstring( f, ", " ); 
      printint( f, cl1 - clauses [ clause1 + CLAUSE_LITERALS ] );
      printstring( f, ", " );
      if( lit1 >= 0 )
      {
         printint( f, lit1 );
         printstring( f, ", " );
      }

      printclause( f, clause2 ); printstring( f, ", " );
      printint( f, cl2 - clauses [ clause2 + CLAUSE_LITERALS ] );
      printstring( f, ", " );
      if( lit2 >= 0 )
      {
         printint( f, lit2 );
         printstring( f, ", " );
      }
      just_printsubst( f, 0 ); printstring( f, ", " );
      just_printsubst( f, 1 ); printstring( f, ").\n\n\n" );
   }
   else
   {
      printstring( f, s );
      printstring( f, ": " );
      printclause( f, result );

      printstring( f, "parent0[" );
      printint( f, cl1 - clauses [ clause1 + CLAUSE_LITERALS ] );
      if( lit1 >= 0 )
      {
         printstring( f, "; " );
         printint( f, lit1 );
      }
      printstring( f, "]: " );
      printclause( f, clause1 );

      printstring( f, "parent1[" );
      printint( f, cl2 - clauses [ clause2 + CLAUSE_LITERALS ] );
      if( lit2 >= 0 )
      {
         printstring( f, "; " );
         printint( f, lit2 );
      }
      printstring( f, "]: " );
      printclause( f, clause2 );

      just_printsubst( f, 0 );
      just_printsubst( f, 1 );
      printstring( f, "\n" );
   }
}


static void printunaryjust( f, s, result, clause, cl1a, cl1b )
bliksemfile *f;
char s [];
int result;
int clause, cl1a, cl1b;
{

   if( cl1a < 0 ) internalproblem( "there should be one parameter at least" );

   if( prologoutput )
   {
      printstring( f, s );
      printstringnnl( f, "(\n");
      printclause( f, result );
      printstring( f, ", " );

      printclause( f, clause );  printstring( f, ", " );
      printint( f, cl1a - clauses [ clause + CLAUSE_LITERALS ] );
      printstring( f, ", " );
      if( cl1b >= 0 )
      {
         printint( f, cl1b - clauses [ clause + CLAUSE_LITERALS ] );
         printstring( f, ", " );
      }
      just_printsubst( f, 0 ); printstring( f, ").\n\n\n" );
   }
   else
   {
      printstring( f, s );
      printstring( f, ": " );
      printclause( f, result );

      printstring( f, "parent0[" );
      printint( f, cl1a - clauses [ clause + CLAUSE_LITERALS ] );
      if( cl1b >= 0 )
      {
         printstring( f, ", " );
         printint( f, cl1b - clauses [ clause + CLAUSE_LITERALS ] );
      }
      printstring( f, "]: " );
      printclause( f, clause );

      just_printsubst( f, 0 );
      printstring( f, "\n" );
   }
}


static void printsubsjust( f, s, result, clause )
bliksemfile *f;
char s [];
int result;
int clause;
{

   if( prologoutput )
   {
      printstring( f, s );
      printstringnnl( f, "(\n" );
      printclause( f, result );
      printstring( f, ", " );
      printclause( f, clause );
      printstring( f, ", " );
      just_printsubst( f, 0 );
      printstring( f, ", " );
      just_printpermutation( f, clause, result );
      printstring( f, " ).\n\n\n" ); 
   }
   else
   {
      printstring( f, s );
      printstring( f, ": " );
      printclause( f, result );

      printstring( f, "parent0: " );
      printclause( f, clause );

      just_printsubst( f, 0 );
      just_printpermutation( f, clause, result );
      printstring( f, "\n" );
   }
}
 

static void printjustification( f, just )
bliksemfile *f;
int just;
{
   int clause1, clause2;
   int cl1, cl2;
   int p;

   switch( justifications [ just ] )
   {
   case JUST_FACTOR:
      p = just + 6;
      readsubstitution( &p );
      readrenaming( &p );

      printunaryjust( f, "factor", justifications [ just + 2 ],
                      justifications [ just + 3 ],
                      justifications [ just + 4 ],
                      justifications [ just + 5 ] );

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_FACTOR" ); 
      break; 
   case JUST_EQFACT:
      p = just + 6;
      readsubstitution( &p );
      readrenaming( &p );

      printunaryjust( f, "eqfact", justifications [ just + 2 ],
                      justifications [ just + 3 ],
                      justifications [ just + 4 ],
                      justifications [ just + 5 ] );

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_EQFACT" ); 
      break; 
   case JUST_PARAMOD:
      p = just + 8; 
      readsubstitution( &p );
      readrenaming( &p );

      printbinaryjust( f, "paramod", justifications [ just + 2 ],
               justifications [ just + 3 ], justifications [ just + 4 ], -1, 
               justifications [ just + 5 ], justifications [ just + 6 ], 
               justifications [ just + 7 ] );  

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_PARAMOD" );
      break; 
   case JUST_RESO:
      p = just + 7;
      readsubstitution( &p );
      readrenaming( &p );

      printbinaryjust( f, "resolution", justifications [ just + 2 ],
               justifications [ just + 3 ], justifications [ just + 4 ], -1,
               justifications [ just + 5 ], justifications [ just + 6 ], -1 );

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_RESO" );
      break;
   case JUST_SUBSUMES:
      p = just + 4;
      readsubstitution( &p );
      readrenaming( &p );
      printsubsjust( f, "subsumption", justifications [ just + 2 ], 
                     justifications [ just + 3 ] );

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_SUBSUMES" );
      break;
   case JUST_EQSWAP:
      p = just + 5;
      readsubstitution( &p );
      readrenaming( &p );

      printunaryjust( f, "eqswap", justifications [ just + 2 ],
                         justifications [ just + 3 ], 
                         justifications [ just + 4 ], -1 );

      if( p != justifications [ just + 1 ] )
         internalproblem( "JUST_EQSWAP" );
      break;  
   case JUST_EQREFL: 
      p = just + 5;
      readsubstitution( &p );
      readrenaming( &p );

      printunaryjust( f, "eqrefl", justifications [ just + 2 ],
                         justifications [ just + 3 ],
                         justifications [ just + 4 ], -1 );

      if( p != justifications [ just + 1 ] ) 
         internalproblem( "JUST_EQREFL" );
      break; 
   default:
      errorwithoutline( "garbage in justifications" );
      break; 
   }
}


static void addtoproof( cl )
int cl;
{
   int i;
   for( i = 0; i < freeproof; i ++ )
   {
      if( proof [i] == cl )
         return;
   }
   proof [ freeproof ++ ] = cl;
   if( freeproof >= MAXPROOF - 4 ) 
      errorwithoutline( "proof is full" );
}


/* justification j occurs in proof.  */

static int justinproof( j )
int j;
{
   int i;

   for( i = 0; i < freeproof; i ++ ) 
   {
      if( proof [i] == justifications [ j + 2 ] ) 
         return 1;
   }
   return 0;
}


static void printproof( f )
bliksemfile *f;
{
   int i, j;
   int freeproof2;

#if 0
   printstring( f, "Rederivation of Goal:\n\n");
#endif

   proof [0] = just_goal;
   freeproof = 1;

   freeproof2 = 0;
   while( freeproof2 < freeproof )
   {
      freeproof2 = freeproof;

      for( j = just_justificationsbase; 
           j < freejustifications; 
           j = justifications [ j + 1 ] )
      {
         if( justinproof( j ))
         {
            switch( justifications [j] )
            {
            case JUST_FACTOR: 
            case JUST_EQFACT:
            case JUST_SUBSUMES:
            case JUST_EQSWAP: 
            case JUST_EQREFL: 
               addtoproof( justifications [ j + 3 ] );
               break;
            case JUST_PARAMOD: 
            case JUST_RESO:
               addtoproof( justifications [ j + 3 ] );
               addtoproof( justifications [ j + 5 ] );
               break;
            default:
               internalproblem( "Don't know justification" ); 
            }
         }
      }
   }

   for( j = just_justificationsbase; 
        j < freejustifications; 
        j = justifications [ j + 1 ] )
   { 
      if( justinproof(j) )
         printjustification( f, j );
   }
}

   
 
/*********************************************************************/
/*        Functions that support the derivation rules.               */
/*********************************************************************/

/* Check if clause is forward subsumed. No fancy discrimination  */
/* trees here.                                                   */

static int just_subsumes( clause1, clause2 )
int clause1;
int clause2;
{

   if( clauses [ clause1 + CLAUSE_LENGTH ] !=
       clauses [ clause2 + CLAUSE_LENGTH ] ) 
      return 0;

   if( subsumes( clause1, clause2 ) /* && subsumes( clause2, clause1 ) */ )
      return 1;
   return 0;
}



static int just_forwardsubsumed( clause, list1, list2 )
int clause;
int list1, list2;
{
   int cl;

#if 0
   printstring( &outputfile, "Forward subsumed ?: " );
   printclause( &outputfile, clause );
#endif

   if( clauses [ clause + CLAUSE_WEIGHT ] > just_maxweight ) 
      return 1;
   if( clauses [ clause + CLAUSE_NRVARS ] > MAXNRVARS - 3 ) 
      return 1;

   for( cl = list1;
        cl != EMPTYCLAUSELIST; 
        cl = clauses [ cl + CLAUSE_NEXTINLIST ] ) 
   {
      if( just_subsumes( cl, clause ))
         return 1;
   }

   for( cl = list2; 
        cl != EMPTYCLAUSELIST;
        cl = clauses [ cl + CLAUSE_NEXTINLIST ] )
   {
      if( just_subsumes( cl, clause ))
         return 1;
   }
   return 0; 
}


/* make a header. The clauses par1 and par2 are parents, when they  */
/* are != EMPTYCLAUSELIST.                                          */


static int just_makeheader( par1, par2 )
int par1, par2;
{
   int g1, g2;
   int result, firstliteral; 

   if( freeclauses + CLAUSE_REST >= maxclauses10 ) 
      more_clauses( );

   result = freeclauses; 
   firstliteral = result + CLAUSE_REST;
   clauses [ result + CLAUSE_NUMBER ] = just_nrkept;

   if( par1 != EMPTYCLAUSELIST )
      g1 = clauses [ par1 + CLAUSE_GENERATION ];
   else
      g1 = -1;

   if( par2 != EMPTYCLAUSELIST )
      g2 = clauses [ par2 + CLAUSE_GENERATION ];
   else
      g2 = -1;
   if( g2 < g1 )
      g1 = g2;

   clauses [ result + CLAUSE_GENERATION ] = g1 + 1;
   clauses [ result + CLAUSE_NEXTINLIST ] = EMPTYCLAUSELIST;
   clauses [ result + CLAUSE_DERIVATION ] = firstliteral;
   clauses [ result + CLAUSE_LITERALS ] = firstliteral;
   clauses [ result + CLAUSE_MAXLITS ] = firstliteral;
   freeclauses = firstliteral;

   return result; 
}


/* Fill in the CLAUSE_END reference and the complexity measures of clause. */

static void just_processnewclause( clause )
int clause;
{
   clauses [ clause + CLAUSE_END ] = freeclauses;

   clauses [ clause + CLAUSE_LENGTH ] = 
         freeclauses - clauses [ clause + CLAUSE_LITERALS ];
   clauses [ clause + CLAUSE_WEIGHT ] = clauseweight( clause ); 
   clauses [ clause + CLAUSE_DEPTH ] = clausedepth( clause ); 
   clauses [ clause + CLAUSE_NRVARS ] = clausenrvars( clause ); 
}


 

/* make copies of the literals clauses [ first .. last ]     */
/* skipping clauses [ toskip ]. origin is necessary for      */
/* finding the right substitution.                           */ 

 
static void just_sideliterals( clause, toskip, origin )
int clause;
int toskip;
int origin; 
{
   int i;
   int t;

   for( i = clauses [ clause + CLAUSE_LITERALS ]; 
        i < clauses [ clause + CLAUSE_END ]; 
        i ++ ) 
   {
      t = ressubst( clauses [i], origin );
      if( i != toskip )
      {
         maketermends( t, freetermspace ); 
         if( freeclauses >= maxclauses10 )
            more_clauses( ); 

         clauses [ freeclauses ++ ] = t; 
      }
      else
      {
         /* It is generated, so that the renaming will be complete. */
         freetermspace = t;
      }
   }
}


/* Return 0 if a non-variable is assigned to (*var, *orig). If a    */
/* variable is assigned to (*var, *orig) then make *var, *orig      */
/* equal to this variable.                                          */

static int endvar( var, orig )
int *var;
int *orig;
{
   int i;
   int a;
restart:
   for( i = 0; i < freesubst; i ++ )
   {
      if( subst [i]. var == *var && subst [i]. varorig == *orig )
      {
         a = termspace [ subst [i]. term ];
         if( a >= 0 )
            return 0;
         *var = a;
         *orig = subst [i]. termorig; 
         goto restart;
      }
   }
   return 1;
}
   
 
/* Return 1 if the substitution is a renaming on orig. */

static int substisrenaming( orig )
int orig;
{
   int var1, varorig1;
   int var2, varorig2;
   int i1, i2;

   for( i1 = 0; i1 < freesubst; i1 ++ )
   {
      if( subst [i1]. varorig == orig )
      {
         var1 = subst [i1]. var;
         varorig1 = subst [i1]. varorig;
         if( endvar( &var1, &varorig1 ))
         {
            if( subst[i1].var != var1 &&
                subst[i1].varorig == varorig1 )
               return 0;

            for( i2 = i1 + 1; i2 < freesubst; i2 ++ )
            if( subst [i2]. varorig == orig )
            {
               var2 = subst [i2]. var;
               varorig2 = subst [i2]. varorig;
               if( endvar( &var2, &varorig2 ))
               {
                  if( var1 == var2 && varorig1 == varorig2 )
                     return 0;
               }
               else
                  return 0;
            }
         }
         else
            return 0; 
      }
   }
   return 1;
}


/***************************************************************/
/*          Factoring.                                         */
/***************************************************************/


static void just_factor( clause, checked, unchecked )
int clause;
int *checked;
int *unchecked;
{
   int cl1, cl2; 
   int thisclause, firstterm, thisjustification;
   int equiv;

#if 0
   printstring( &outputfile, "attempting factoring ");
   printclause( &outputfile, clause );
#endif

   for( cl1 = clauses [ clause + CLAUSE_LITERALS ];
        cl1 < clauses [ clause + CLAUSE_END ];
        cl1 ++ )
   {
      for( cl2 = cl1 + 1; 
           cl2 < clauses [ clause + CLAUSE_END ];
           cl2 ++ ) 
      {
         freesubst = 0;
         if( unifyeqrewr( clauses [ cl1 ], 0, clauses [ cl2 ], 0 ))
         {
            firstterm = freetermspace;
            thisclause = just_makeheader( clause, EMPTYCLAUSELIST );
            freerenaming = 0;
    
            just_sideliterals( clause, cl2, 0 );
            just_processnewclause( thisclause );

            if( just_forwardsubsumed( thisclause, *unchecked, *checked ))
            {
               freetermspace = firstterm;
               freeclauses = thisclause;
            }
            else
            {
               clauses [ thisclause + CLAUSE_GENERATION ] = 
                  clauses [ clause + CLAUSE_GENERATION ];

               clauses [ thisclause + CLAUSE_NEXTINLIST ] = *unchecked;
               *unchecked = thisclause;
               just_nrkept ++ ;

               if( freejustifications >= maxjustifications10 )
                  more_justifications( ); 
               thisjustification = freejustifications; 
               justifications [ thisjustification + 0 ] = JUST_FACTOR;
               justifications [ thisjustification + 2 ] = thisclause;
               justifications [ thisjustification + 3 ] = clause;
               justifications [ thisjustification + 4 ] = cl1;
               justifications [ thisjustification + 5 ] = cl2;
               freejustifications = thisjustification + 6; 
               storesubstitution( );
               storerenaming( );  
               justifications [ thisjustification + 1 ] = freejustifications; 
            }
         }
      }
   }
}

/* Close the set clauselist under factoring.               */

static void close_factoring( clauselist )
int *clauselist;
{
   int given, unchecked;

#if 0
   printstring( &outputfile, "closing under factoring: " );
   printclauselist( &outputfile, *clauselist );
   printstring( &outputfile, "\n" );
#endif

   unchecked = *clauselist;
   *clauselist = EMPTYCLAUSELIST;

   while( unchecked != EMPTYCLAUSELIST )
   {
      given = unchecked;
      unchecked = clauses [ unchecked + CLAUSE_NEXTINLIST ];
      clauses [ given + CLAUSE_NEXTINLIST ] = *clauselist;
      *clauselist = given;
  
      just_factor( given, clauselist, &unchecked );
   }

   /* It might be a good idea to apply backward subsumption here.   */

}

/****************************************************************/
/*                    The derivation rules:                     */
/****************************************************************/


/* Try to swap equalities/rewrites, positive or negative.       */

static void just_swapeq( clause, checked, unchecked )
int clause;
int *checked;
int *unchecked;
{
   int cl;
   int thisclause, firstterm, thisjustification; 
   int at, lit;
   int a, l;
   int i;

   for( cl = clauses [ clause + CLAUSE_LITERALS ];
        cl < clauses [ clause + CLAUSE_END ]; 
        cl ++ )
   {
      lit = clauses [ cl ];
      if( termspace [ lit ] == OP_NOT ) 
         at = lit + 1;
      else
         at = lit;

      if( termspace [ at ] == OP_REWRITES || 
          termspace [ at ] == OP_EQUAL ) 
      {
         freesubst = 0;
         firstterm = freetermspace;
         thisclause = just_makeheader( clause, EMPTYCLAUSELIST );
         freerenaming = 0;

         l = ressubst( clauses [ cl ], 0 );
         maketermends( l, freetermspace ); 
         if( termspace [l] == OP_NOT )
            a = l + 1;
         else
            a = l;
         exchangearguments( a );
         clauses [ freeclauses ++ ] = l;

         just_sideliterals( clause, cl, 0 );

         just_processnewclause( thisclause ); 
         if( just_forwardsubsumed( thisclause, *unchecked, *checked ))
         {
            freetermspace = firstterm;
            freeclauses = thisclause;
         }
         else
         {
            clauses [ thisclause + CLAUSE_GENERATION ] = 
               clauses [ clause + CLAUSE_GENERATION ];

            clauses [ thisclause + CLAUSE_NEXTINLIST ] = *unchecked;
            *unchecked = thisclause; 
            just_nrkept ++ ;

            if( freejustifications >= maxjustifications10 ) 
               more_justifications( ); 
 
            thisjustification = freejustifications;  
            justifications [ thisjustification + 0 ] = JUST_EQSWAP;
            justifications [ thisjustification + 2 ] = thisclause;
            justifications [ thisjustification + 3 ] = clause;
            justifications [ thisjustification + 4 ] = cl;
            freejustifications = thisjustification + 5;
            storesubstitution( );
            storerenaming( );
            justifications [ thisjustification + 1 ] = freejustifications;
         }
      }
   }
}


/* Close clauselist under equality swapping.                    */

static void close_eqswap( clauselist )
int *clauselist;
{
   int given, unchecked;

#if 0
   printstring( &outputfile, "closing under equality swapping: " );
   printclauselist( &outputfile, *clauselist );
   printstring( &outputfile, "\n" );
#endif

   unchecked = *clauselist;
   *clauselist = EMPTYCLAUSELIST;

   while( unchecked != EMPTYCLAUSELIST )
   {
      given = unchecked;
      unchecked = clauses [ unchecked + CLAUSE_NEXTINLIST ];
      clauses [ given + CLAUSE_NEXTINLIST ] = *clauselist;
      *clauselist = given;
  
      just_swapeq( given, clauselist, &unchecked );
   }

   /* Would a backward subsumption check be a good idea?   */

}

/****************************************************************/
/*                    The derivation rules:                     */
/****************************************************************/


/* Try to apply the eqrefl rule. If simp = 1, the substitution   */ 
/* should be a renaming.                                         */

static void just_eqrefl( clause, results, simp )
int clause;
int *results;
int simp;
{
   int cl;
   int thisclause, firstterm, thisjustification;
   int lit;
   int arg1, arg2;
   int equiv;

#if 0
   printstring( &outputfile, "attempting eqrefl on " );
   printclause( &outputfile, clause );
   printstring( &outputfile, "\n" );
#endif
   *results = EMPTYCLAUSELIST;

   for( cl = clauses [ clause + CLAUSE_LITERALS ];
        cl < clauses [ clause + CLAUSE_END ];
        cl ++ ) 
   {
      lit = clauses [ cl ];
      if( termspace [ lit ] == OP_NOT && 
          ( termspace [ lit + 1 ] == OP_EQUAL || 
            termspace [ lit + 1 ] == OP_REWRITES ))
      {
         arg1 = lit + 2;
         arg2 = termends [ lit + 2 ];

         freesubst = 0;
         if( unify( arg1, 0, arg2, 0 ))
         {
            firstterm = freetermspace;
            thisclause = just_makeheader( clause, EMPTYCLAUSELIST );
            freerenaming = 0;
   
            just_sideliterals( clause, cl, 0 );
            just_processnewclause( thisclause );

            equiv = 1;
            if( simp == 1 && !substisrenaming(0) )
               equiv = 0;

            if( !equiv || just_forwardsubsumed( thisclause, 
                   *results, EMPTYCLAUSELIST ))
            {
               freetermspace = firstterm;
               freeclauses = thisclause;
            }
            else
            {
               clauses [ thisclause + CLAUSE_NEXTINLIST ] = *results;
               *results = thisclause;
               just_nrkept ++ ;

               if( freejustifications >= maxjustifications10 ) 
                  more_justifications( );
               thisjustification = freejustifications;
               justifications [ thisjustification + 0 ] = JUST_EQREFL;
               justifications [ thisjustification + 2 ] = thisclause;
               justifications [ thisjustification + 3 ] = clause;
               justifications [ thisjustification + 4 ] = cl; 
               freejustifications = thisjustification + 5; 
               storesubstitution( );
               storerenaming( );
               justifications [ thisjustification + 1 ] = freejustifications;  
            }
         }
      }
   }
}



/* Try to make an equality factoring.                  */


static void just_eqfact( clause, results )
int clause;
int *results;
{
   int cl1, cl2;
   int lit1, lit2;
   int thisclause, firstterm, thisjustification;
   int t1, t1a, t2, t3;
  
#if 0
   printstring( &outputfile, "attempting eqfact on ");
   printclause( &outputfile, clause );
   printstring( &outputfile, "\n");
#endif

   *results = EMPTYCLAUSELIST;

   for( cl1 = clauses [ clause + CLAUSE_MAXLITS ];
        cl1 < clauses [ clause + CLAUSE_END ];
        cl1 ++ )
   {
      lit1 = clauses [ cl1 ];
      if( termspace [ lit1 ] == OP_EQUAL || 
          termspace [ lit1 ] == OP_REWRITES )
      {
         for( cl2 = clauses [ clause + CLAUSE_LITERALS ]; 
              cl2 < clauses [ clause + CLAUSE_END ];
              cl2 ++ ) 
         {
            if( cl1 != cl2 )
            {
               lit2 = clauses [ cl2 ];
               if( termspace [ lit2 ] == OP_EQUAL ||
                   termspace [ lit2 ] == OP_REWRITES ) 
               {
                  t1 = lit1 + 1;
                  t2 = termends [ t1 ];
                  t1a = lit2 + 1;
                  t3 = termends [ t1a ];

                  /* From cl1: t1 = t2, cl2: t1 = t3 we must produce   */
                  /* ! t2 = t3, t1 = t3.                               */
 
                  freesubst = 0;
                  if( unifyeqrewr( t1, 0, t1a, 0 ))
                  {
                     firstterm = freetermspace;
                     thisclause = just_makeheader( clause, EMPTYCLAUSELIST );
                     freerenaming = 0;
                     
                     clauses [ freeclauses ++ ] = freetermspace;
                     termspace [ freetermspace ++ ] = OP_NOT;
                     termspace [ freetermspace ++ ] = OP_EQUAL;
                     (void) ressubst( t2, 0 );
                     (void) ressubst( t3, 0 );   
                     maketermends( firstterm, freetermspace );
 
                     just_sideliterals( clause, cl1, 0 );
                     just_processnewclause( thisclause );

                     if( just_forwardsubsumed( thisclause, 
                        *results, EMPTYCLAUSELIST ))
                     {
                        freetermspace = firstterm;
                        freeclauses = thisclause;
                     }
                     else
                     {
                        clauses [ thisclause + CLAUSE_NEXTINLIST ] = *results;
                        *results = thisclause;
                        just_nrkept ++ ;

                        if( freejustifications >= maxjustifications10 )
                           more_justifications( );  
                        thisjustification = freejustifications; 
                        justifications [ thisjustification + 0 ] = JUST_EQFACT;
                        justifications [ thisjustification + 2 ] = thisclause;
                        justifications [ thisjustification + 3 ] = clause;
                        justifications [ thisjustification + 4 ] = cl1;
                        justifications [ thisjustification + 5 ] = cl2;
                        freejustifications = thisjustification + 6; 
                        storesubstitution( );
                        storerenaming( );  
                        justifications [ thisjustification + 1 ] = 
                              freejustifications; 
                     }
                  }
               }
            }
         }
      }
   }
}


/* Return the nth occurrence of subterm in term.  */
 
static int nthoccurrence( term, n, subterm )
int term;
int n;
int subterm;
{
   int t;
 
   if( n == 0 ) 
      internalproblem( "n=0" );
 
   for( t = term; t < termends [ term ]; t ++ )
   {
      if( termequal( t, subterm ))
      {
         n -- ;
         if( n == 0 )
            return t;
      }
   }
   return t;
}
 


/* Produce all 1-step paramodulations from from into into.    */
/* Results are added to *unchecked. If simp = 1, the          */
/* the substitution must be a renaming on into.               */


static void just_paramod2( from, into, checked, unchecked, simp )
int from;
int into;
int *checked;
int *unchecked;
int simp;
{
   int cl1, cl2;
   int eq; 
   int arg1, arg2;
   int arg1subst, arg2subst; 
   int lit2, at2;
   int t2;
   int thisclause, firstterm, thisjustification;
   int n;
   int x1, x2; 
   int clauseterm, repl;

#if 0
   printstring( &outputfile, "attempting 1-step paramodulation from ");
   printclause( &outputfile, from );
   printstring( &outputfile, "   into\n" );
   printclause( &outputfile, into );
   printstring( &outputfile, "\n" );
#endif

   for( cl1 = clauses [ from + CLAUSE_LITERALS ];
        cl1 < clauses [ from + CLAUSE_END ];
        cl1 ++ ) 
   {
      eq = clauses [ cl1 ];
      if( termspace [ eq ] == OP_REWRITES || termspace [ eq ] == OP_EQUAL ) 
      {
         arg1 = eq + 1; 
         arg2 = termends [ eq + 1 ];
         for( cl2 = clauses [ into + CLAUSE_LITERALS ];
              cl2 < clauses [ into + CLAUSE_END ];
              cl2 ++ ) 
         { 
            lit2 = clauses [ cl2 ];
            if( termspace [ lit2 ] == OP_NOT ) 
               at2 = lit2 + 1; 
            else
               at2 = lit2;
            for( t2 = at2 + 1; t2 < termends [ at2 ]; t2 ++ ) 
            {
               if( termspace [ t2 ] >= 0 )  /* Not into variable. */
               {
                  freesubst = 0;
                  if( unify( arg1, 0, t2, 1 )) 
                  {
                     if( ( termapriori || termaposteriori ) && 
                           termcompare( arg1, arg2 ) == COMP_LESS )
                        goto failure;
                     if( simp && !substisrenaming(1))
                        goto failure;
 
                     n = 1;
                     while( n >= 0 ) 
                     {
                        firstterm = freetermspace;
                        thisclause = just_makeheader( from, into ); 
 
                        freerenaming = 0; 
                        clauseterm = ressubst( lit2, 1 );
                        maketermends( clauseterm, freetermspace );
                        arg1subst = ressubst( arg1, 0 );
                        maketermends( arg1subst, freetermspace );
                        arg2subst = ressubst( arg2, 0 );
                        maketermends( arg2subst, freetermspace );

                        repl = nthoccurrence( clauseterm, n, arg1subst ); 
 
                        /* We have:                                     */
                        /* clauseterm: Into literal after substitution. */
                        /* arg1subst: arg1 after substitution.          */
                        /* arg2subst: arg2 after substitution.          */
                        /* repl: A position in clauseterm where         */
                        /* arg1subst occurs.                            */

                        if( repl != termends [ clauseterm ] )
                        {
                           x2 = replaceposition( clauseterm, 
                                                 repl, 
                                                 arg2subst );
                           x1 = clauseterm;
                           while( x2 < freetermspace ) 
                              termspace [ x1 ++ ] = termspace [ x2 ++ ];
                           freetermspace = x1;
                           maketermends( clauseterm, freetermspace );
 
                           if( freeclauses >= maxclauses10 ) 
                              more_clauses( );

                           clauses [ freeclauses ++ ] = clauseterm;

                           just_sideliterals( from, cl1, 0 );
                           just_sideliterals( into, cl2, 1 );
                           just_processnewclause( thisclause );

                           if( just_forwardsubsumed( thisclause,
                                    *checked, *unchecked ))
                           {
                              freetermspace = firstterm;
                              freeclauses = thisclause;
                           }
                           else
                           {
                              clauses [ thisclause + CLAUSE_NEXTINLIST ] = 
                                 *unchecked;
                              *unchecked = thisclause;
                              just_nrkept ++ ;
 
                              if(freejustifications >= maxjustifications10 ) 
                                 more_justifications( );  
                              thisjustification = freejustifications;
                              justifications [ thisjustification ] = 
                                    JUST_PARAMOD;
                              justifications [ thisjustification + 2 ] = 
                                    thisclause;
                              justifications [ thisjustification + 3 ] = 
                                    from;
                              justifications [ thisjustification + 4 ] = cl1;
                              justifications [ thisjustification + 5 ] = 
                                    into;
                              justifications [ thisjustification + 6 ] = cl2;
                              justifications [ thisjustification + 7 ] = 
                                    repl - clauseterm;
                              freejustifications = thisjustification + 8; 
                              storesubstitution( );
                              storerenaming( );
                              justifications [ thisjustification + 1 ] =
                                 freejustifications; 

                           }
                           n ++; 
                        }
                        else
                        {
                           freetermspace = firstterm;
                           freeclauses = thisclause;
                           n = -1;
                        }
                     }
                  }
               failure:
                  ;
               }
            }
         }
      }
   }
}


/* Produce paramodulants in *results. If simp = 1, then    */
/* the substitution must be a renaming on into.            */

static void just_paramod( from, into, results, simp )
int from;
int into;
int *results;
int simp;
{
   int given, unchecked;
   int cnt;

   *results = EMPTYCLAUSELIST;
   unchecked = EMPTYCLAUSELIST;

   just_paramod2( from, into, results, &unchecked, simp );
   cnt = 0;
   while( unchecked != EMPTYCLAUSELIST && cnt < MAXPARAMOD )
   {
      given = unchecked;
      unchecked = clauses [ unchecked + CLAUSE_NEXTINLIST ];
      clauses [ given + CLAUSE_NEXTINLIST ] = *results;
      *results = given;

      just_paramod2( from, given, results, &unchecked, 1 );
      cnt ++ ;
   }
   while( unchecked != EMPTYCLAUSELIST )
   {
      given = unchecked;
      unchecked = clauses [ unchecked + CLAUSE_NEXTINLIST ];
      clauses [ given + CLAUSE_NEXTINLIST ] = *results;
      *results = given;
   }
   return;
}


/* simp tells on which of negative/positive the resolvent should be a   */
/* simplification, -1 for negative, 1 for positive, 0 for nowhere.      */

static void just_reso( negative, positive, results, simp ) 
int negative;
int positive; 
int *results;
int simp;
{
   int cl1, cl2;
   int lit1, lit2;
   int thisclause, firstterm, thisjustification; 
   int equiv;

#if 0
   printstring( &outputfile, "attempting resolution:\n");
   printclause( &outputfile, negative );
   printclause( &outputfile, positive );
   printstring( &outputfile, "(" );
   printint( &outputfile, simp );
   printstring( &outputfile, ")\n" );
#endif

   *results = EMPTYCLAUSELIST;
 
   for( cl1 = clauses [ negative + CLAUSE_LITERALS ]; 
        cl1 < clauses [ negative + CLAUSE_END ]; 
        cl1 ++ )
   {
      if( termspace [ clauses [ cl1 ]] == OP_NOT ) 
      {
         lit1 = clauses [ cl1 ] + 1;
         for( cl2 = clauses [ positive + CLAUSE_LITERALS ]; 
              cl2 < clauses [ positive + CLAUSE_END ];
              cl2 ++ ) 
         {
            if( termspace [ clauses [ cl2 ]] != OP_NOT )
            {
               lit2 = clauses [ cl2 ];

               freesubst = 0;
               if( unifyeqrewr( lit1, 0, lit2, 1 )) 
               {
                  firstterm = freetermspace;
                  thisclause = just_makeheader( negative, positive );
                  freerenaming = 0;
 
                  just_sideliterals( negative, cl1, 0 );
                  just_sideliterals( positive, cl2, 1 );  
                 
                  just_processnewclause( thisclause ); 

                  /* Check the simplification conditions.   */

                  equiv = 1;  
                  if( simp == -1 && !substisrenaming( 0 ))
                     equiv = 0;
                  if( simp == 1 && !substisrenaming( 1 ))
                     equiv = 0; 
 
                  if( !equiv || just_forwardsubsumed( thisclause, 
                                        *results, EMPTYCLAUSELIST ))
                  {
                     freetermspace = firstterm;
                     freeclauses = thisclause;
                  }
                  else
                  {
                     clauses [ thisclause + CLAUSE_NEXTINLIST ] = *results;
                     *results = thisclause;
                     just_nrkept ++ ;

                     if( freejustifications >= maxjustifications10 ) 
                        more_justifications( ); 
                     thisjustification = freejustifications;
                     justifications [ thisjustification ] = JUST_RESO;
                     justifications [ thisjustification + 2 ] = thisclause;
                     justifications [ thisjustification + 3 ] = negative;
                     justifications [ thisjustification + 4 ] = cl1;
                     justifications [ thisjustification + 5 ] = positive;
                     justifications [ thisjustification + 6 ] = cl2;
                     freejustifications = thisjustification + 7;
                     storesubstitution( );
                     storerenaming( );
                     justifications [ thisjustification + 1 ] = 
                        freejustifications; 
                  }
               }
            }
         }
      }
   }
}


/* Check whether or not clauses [ from1 .. from2 ] subsumes  */
/* clauses [ into1 .. into2 ].                               */


static int simplesubs( from1, from2, into1, into2 )
int from1;
int from2;
int into1;
int into2;
{
   int i;
   int matchingbase;

   if( from1 == from2 )
      return 1;
   else
   {
      matchingbase = freematching;
      for( i = into1; i < into2; i ++ )
      {
         if( matcheqrewr( clauses [ from1 ], clauses [i] ))
         {
            if( simplesubs( from1 + 1, from2, into1, into2 ))
               return 1;
         }
         freematching = matchingbase;
      }
      return 0; 
   }
}


/* collect all the variables of clause into the renaming, renaming   */
/* them into themselves.                                             */


static void collectvars( clause )
int clause;
{
   int cl;
   int t, tend;
   int i;
   int minvar;

   minvar = 0; 
   for( cl = clauses [ clause + CLAUSE_LITERALS ];
        cl < clauses [ clause + CLAUSE_END ];
        cl ++ )
   {
      t = clauses [ cl ];
      tend = termends [t];
      while( t < tend )
      {
         if( termspace [t] < minvar )
            minvar = termspace [t];
         t ++ ; 
      }
   }
   freerenaming = -minvar;

   for( i = 0; i < freerenaming; i ++ ) 
   {
      renaming [i]. var = - ( 1 + i );
      renaming [i]. varorig = 1;
   }
}


/**********************************************************/
/*               check if clause is equal to just_goal.   */
/**********************************************************/


static int just_isgoal( clause )
int clause;
{
   int thisjustification; 

   if( clauses [ clause + CLAUSE_LENGTH ] != 
       clauses [ just_goal + CLAUSE_LENGTH ] ) 
      return 0;

   freematching = 0;
   if( simplesubs( 
      clauses [ clause + CLAUSE_LITERALS ], 
      clauses [ clause + CLAUSE_END ], 
      clauses [ just_goal + CLAUSE_LITERALS ],
      clauses [ just_goal + CLAUSE_END ] ))
   {
      if( freejustifications >= maxjustifications10 ) 
         more_justifications( ); 
      thisjustification = freejustifications; 
      justifications [ thisjustification ] = JUST_SUBSUMES;
      justifications [ thisjustification + 2 ] = just_goal;
      justifications [ thisjustification + 3 ] = clause; 
      freejustifications = thisjustification + 4;
      storematching( );    

      /* We need a renaming here for the symmetry. */

      collectvars( just_goal );
      storerenaming( );

      justifications [ thisjustification + 1 ] = freejustifications;
      return 1;
   }
   return 0; 
}




/*******************************************************************/
/*  Functions for reproducing the real initial clauses.            */
/*  We are really redoing a part of rules.c here, but we want to   */
/*  keep rules.c clean.                                            */
/*******************************************************************/



/* Translate the list listterm into a clause.    */

static void makeliterals( listterm )
int listterm;
{
   int res;
 
   while( termspace [ listterm ] != OP_NIL )
   {
      listterm ++ ;
      res = copyterm( listterm );
      clauses [ freeclauses ++ ] = res;
      if( freeclauses >= maxclauses10 ) 
         more_clauses( ); 
      listterm = termends [ listterm ];
   }
}
 

/* Make an initial clause from term. NO processing takes place. */

static void just_initial( term ) 
int term;
{
   int thisclause; 

   thisclause = freeclauses;
 
   /* make the header:  */

   thisclause = just_makeheader( EMPTYCLAUSELIST, EMPTYCLAUSELIST ); 
 
   if( termspace [ term ] != OP_NIL && termspace [ term ] != OP_DOT ) 
      internalproblem( "Dot expected" );
   makeliterals( term );

   just_processnewclause( thisclause ); 
   just_nrkept ++ ; 
   return;
}



/**********************************************************/
/*       Try to apply a side rule.                        */
/**********************************************************/



static void just_simprule( listonlevel, level )
int listonlevel;
int level;
{
   int freejustificationsbase;
   int freeclausesbase;
   int freetermspacebase;

   int par1, par2;
   int dir;
   int nextgen;

#if 0
   printstring( &outputfile, "Applying simplification rule at level ");
   printint( &outputfile, level );
   printstring( &outputfile, " on clauselist " ); 
   printclauselist( &outputfile, listonlevel );
   printstring( &outputfile, "\n\n" );
#endif
   if( level > freeparents )
      internalproblem( "level incorrect" );

   if( level == freeparents )
   {
      close_eqswap( &listonlevel );
      close_factoring( &listonlevel );
      while( listonlevel != EMPTYCLAUSELIST )
      {
         if( just_isgoal( listonlevel ))
         {
            printproof( &outputfile );
            longjmp( just_success, 1 );
         }
         listonlevel = clauses [ listonlevel + CLAUSE_NEXTINLIST ];
      }
      return;
   }
   if( parents [ level ] >= 0 )
      internalproblem( "level incorrect" );

   close_factoring( &listonlevel );

   switch( parents [ level ] )
   {
   case SIMP_RES:
   case SIMP_DEMOD:
      for( par1 = listonlevel;
           par1 != EMPTYCLAUSELIST;
           par1 = clauses [ par1 + CLAUSE_NEXTINLIST ] )
      {
         for( par2 = parents [ level + 1 ];
              par2 != EMPTYCLAUSELIST; 
              par2 = clauses [ par2 + CLAUSE_NEXTINLIST ] )
         {
            for( dir = 0; dir < 2; dir ++ )
            {
               freejustificationsbase = freejustifications;
               freeclausesbase = freeclauses;
               freetermspacebase = freetermspace;

               if( parents [ level ] == SIMP_RES )
               {
                  if( dir == 0 )
                     just_reso( par1, par2, &nextgen, -1 );
                  else
                     just_reso( par2, par1, &nextgen, 1 );
               }

               if( parents [ level ] == SIMP_DEMOD )
               {
                  if( dir == 0 )
                     just_paramod( par2, par1, &nextgen, 1 );
                  else
                     nextgen = EMPTYCLAUSELIST;
               }

               just_simprule( nextgen, level + 2 );

               freejustifications = freejustificationsbase;
               freeclauses = freeclausesbase;
               freetermspace = freetermspacebase;
            } 
         }
      }
      break;  
   case SIMP_FACTOR:
      just_simprule( listonlevel, level + 1 );
      break; 
   case SIMP_EQREFL:
      for( par1 = listonlevel; 
           par1 != EMPTYCLAUSELIST; 
           par1 = clauses [ par1 + CLAUSE_NEXTINLIST ] )
      {
         freejustificationsbase = freejustifications;
         freeclausesbase = freeclauses;
         freetermspacebase = freetermspace;

         if( parents [ level ] == SIMP_EQREFL )
            just_eqrefl( par1, &nextgen, 1 );

         just_simprule( nextgen, level + 1 );     

         freejustifications = freejustificationsbase;
         freeclauses = freeclausesbase;
         freetermspace = freetermspacebase;
      }
      break;         
   default:
      printf("unknown simplifier %d\n", parents [ level ] );
      internalproblem( "unknown simplifier" );
   } 
}





/* Try to apply the main rule.   */

static int just_mainrule( )
{
   int freejustificationsbase;
   int freeclausesbase;
   int freetermspacebase;

   int par1, par2;
   int dir;
   int nextgen;

#if 0
   printstring( &outputfile, "Applying main rule on:\n" );
   printstring( &outputfile, "Parent1: ");
   printclauselist( &outputfile, just_parent1 );
   printstring( &outputfile, "\n" );
   printstring( &outputfile, "Parent2: ");
   printclauselist( &outputfile, just_parent2 );
   printstring( &outputfile, "\n" );
#endif

   if( setjmp( just_success ))
      return 1;

   switch( parents [0] )
   {
   case RULE_RES:
   case RULE_PARAMOD:
      for( par1 = just_parent1;
           par1 != EMPTYCLAUSELIST; 
           par1 = clauses [ par1 + CLAUSE_NEXTINLIST ] )
      {
         for( par2 = just_parent2;
              par2 != EMPTYCLAUSELIST;
              par2 = clauses [ par2 + CLAUSE_NEXTINLIST ] )
         {
            for( dir = 0; dir < 2; dir ++ )
            {
               freejustificationsbase = freejustifications;
               freeclausesbase = freeclauses;
               freetermspacebase = freetermspace;

               if( parents [0] == RULE_RES )
               {
                  if( dir == 0 )
                     just_reso( par1, par2, &nextgen, 0 );
                  else
                     just_reso( par2, par1, &nextgen, 0 ); 
               }
               if( parents [0] == RULE_PARAMOD )
               {
                  if( dir == 0 )
                     just_paramod( par1, par2, &nextgen, 0 );
                  else
                     just_paramod( par2, par1, &nextgen, 0 );
               }
               just_simprule( nextgen, 3 );

               freejustifications = freejustificationsbase;
               freeclauses = freeclausesbase;
               freetermspace = freetermspacebase;
            }
         }
      }
      break;
   case RULE_INITIAL:
      just_simprule( just_parent1, 1 );
      break;
   case RULE_FACTOR:
   case RULE_RESIMP:
      just_simprule( just_parent1, 2 );
      break;
   case RULE_EQREFL:
   case RULE_EQFACT:
      for( par1 = just_parent1;
           par1 != EMPTYCLAUSELIST;
           par1 = clauses [ par1 + CLAUSE_NEXTINLIST ] )
      {
         freejustificationsbase = freejustifications;
         freeclausesbase = freeclauses;
         freetermspacebase = freetermspace;

         if( parents [0] == RULE_EQREFL )
            just_eqrefl( par1, &nextgen, 0 );
         if( parents [0] == RULE_EQFACT )
            just_eqfact( par1, &nextgen );

         just_simprule( nextgen, 2 );
         freejustifications = freejustificationsbase;
         freeclauses = freeclausesbase;
         freetermspace = freetermspacebase;
      }
      break;
   default:
      internalproblem( "unknown rule\n" );
   }
   return 0;
}


/* return 1 if clause can be successfully derived from its parents,   */
/* together with its optional hiddenparent. If hiddenparent !=        */
/* EMPTYCLAUSELIST, there is no hidden parent.                        */

static int rederive( hiddenparent, clause )
int hiddenparent;
int clause;
{
   int i;
   int sideparents;

   just_goal = clause;

   freeparents = 0;
   for( i = clauses [ just_goal + CLAUSE_DERIVATION ]; 
        i < clauses [ just_goal + CLAUSE_LITERALS ]; 
        i ++ )
   {
      parents [ freeparents ++ ] = clauses [i];
   }
   if( freeparents == 0 )
      internalproblem( "empty parent list" );

   /* We close all parents under equality swapping.      */

   for( i = 0; i < freeparents; i ++ )
   {
      if( parents [i] >= 0 )
      {
         clauses [ parents [i] + CLAUSE_NEXTINLIST ] = EMPTYCLAUSELIST;
         close_eqswap( & parents [i] );
      }
   }

   if( hiddenparent != EMPTYCLAUSELIST )
      clauses [ hiddenparent + CLAUSE_NEXTINLIST ] = EMPTYCLAUSELIST;
   close_eqswap( &hiddenparent );
 
   switch( parents [0] )
   {
   case RULE_RES:
   case RULE_PARAMOD:
      just_parent1 = parents [1];
      just_parent2 = parents [2];
      sideparents = 3;
      break;
   case RULE_EQREFL:
   case RULE_EQFACT:
   case RULE_FACTOR:
   case RULE_RESIMP:
      just_parent1 = parents [1];
      just_parent2 = EMPTYCLAUSELIST;
      sideparents = 2;
      break;
   case RULE_INITIAL:
      if( hiddenparent == EMPTYCLAUSELIST )
         internalproblem( "there should be a hidden parent" );
      just_parent1 = hiddenparent;
      just_parent2 = EMPTYCLAUSELIST;
      sideparents = 1;
      break;
   default:
      internalproblem( "don't know this rule" );
   } 

#if 0
   printstring( &outputfile, "======================================\n"); 
   printstring( &outputfile, "Attempting Rederivation of:\n\n" ); 
   printclause( &outputfile, just_goal );

   if( just_parent1 != EMPTYCLAUSELIST )
   { 
      printstring( &outputfile, "Parent1:\n");
      printclauselist( &outputfile, just_parent1 );
      printstring( &outputfile, "\n" );
   }
   if( just_parent2 != EMPTYCLAUSELIST )
   {
      printstring( &outputfile, "Parent2:\n");
      printclauselist( &outputfile, just_parent2 );
      printstring( &outputfile, "\n" );
   }
   printstring( &outputfile, "Side parents:\n\n");
   for( i = sideparents; i < freeparents; i ++ )
   {
      if( parents [i] >= 0 )
      {
         printclauselist( &outputfile, parents [i] );
         printstring( &outputfile, "\n");  
      }
   }
   printstring( &outputfile, "\n" );
#endif

   i = just_mainrule( );
   return i;
}


void checkproof( startoftranslation, endoftranslation, listproof )
int startoftranslation;
int endoftranslation; 
int listproof;
{
   int freetermspacebase, freeclausesbase;
   int t;
   int i; 

   substype = SUBS_STANDARD;

   freetermspacebase = freetermspace;
   freeclausesbase = freeclauses;
   just_justificationsbase = freejustifications;
   just_maxweight = MAXWEIGHT;
   just_nrkept = nrkept + 1;

   /* Reproduce initial clauses without any processing.   */
   
   just_initials1 = freeclauses;
   for( t = startoftranslation; t < endoftranslation; t = termends [t] )
      just_initial(t);
   just_initials2 = freeclauses;

   if( prologoutput )
   {
      printstring( &outputfile, "% ABCDEFGHIJKLMNOPQRSTUVWXYZ\n\n");

      printstring( &outputfile, "initialclauses(\n" );
    
      if( just_initials1 == just_initials2 )
         printstring( &outputfile, "[] " ); 
      else
      {
         for( t = just_initials1; t < just_initials2;
              t = clauses [ t + CLAUSE_END ] )
         {
            if( t == just_initials1 )
               printstring( &outputfile, "[ ");
            else
               printstring( &outputfile, ", ");
            printclause( &outputfile, t );
         }
         printstring( &outputfile, "] ");
      } 
      printstring( &outputfile, ").\n\n" );
   }
   else
   {
      printstring( &outputfile, "\nUnprocessed initial clauses:\n\n");
      for( t = just_initials1; t < just_initials2;
           t = clauses [ t + CLAUSE_END ] ) 
      {
         printclause( &outputfile, t );
      }
   }

   printstring( &outputfile, "\n\n" );
   if( !prologoutput )
      printstring( &outputfile, "Total Proof:\n\n" );

   /* Redo the proof. Below t the prooflist is destroyed.  */

   for( t = listproof; t != EMPTYCLAUSELIST; 
        t = clauses [ t + CLAUSE_NEXTINLIST ] ) 
   {
      if( clauses [ clauses [ t + CLAUSE_DERIVATION ]] == RULE_INITIAL ) 
      {
         for( i = just_initials1; i < just_initials2; 
              i = clauses [ i + CLAUSE_END ] ) 
         {
            freejustifications = just_justificationsbase; 
            if( rederive( i, t ))
               goto ok;
         }
         goto ohoh; 
      }
      else
      {
         freejustifications = just_justificationsbase;
         if( rederive( EMPTYCLAUSELIST, t ))
            goto ok; 
         goto ohoh; 
      }
   ok:
#if 0
      printstring( &outputfile, "goal rederived!\n\n"); 
#endif
      ;
   }

   if( prologoutput )
   {
      printstring( &outputfile, "end.\n\n" );
      printstring( &outputfile, "% ABCDEFGHIJKLMNOPQRSTUVWXYZ\n\n" );
   }
   else
      printstring( &outputfile, "proof check complete!\n\n" );
   
   freejustifications = just_justificationsbase;
   freeclauses = freeclausesbase;
   freetermspace = freetermspacebase;
 
   return;

ohoh:
   printstring( &outputfile, "\n==> " );
   printclause( &outputfile, t );
   printstring( &outputfile, "\n" );
   internalproblem( "COULD NOT REDERIVE GOAL" );  

   return;  
}



